/******************************************************************************/
/*                                                                            */
/* Module     : gvibBindParm.i                                                */
/*                                                                            */
/* Description: This module is the SWIG source file containing all of the     */
/*              C code associated with binding the input parameter columns.   */
/*              These functions are called from the IBCursorBindParm function */
/*              in the gvibCursor.i file.                                     */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Copyright (c) 2000 Going Virtual, LLC. All rights reserved.                */
/*                                                                            */
/* This library is free software; you can redistribute it and/or              */
/* modify it under the terms of the GNU Lesser General Public                 */
/* License as published by the Free Software Foundation; either               */
/* version 2.1 of the License, or (at your option) any later version.         */
/*                                                                            */
/* This library is distributed in the hope that it will be useful,            */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of             */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          */
/* Lesser General Public License for more details.                            */
/*                                                                            */
/* You should have received a copy of the GNU Lesser General Public           */
/* License along with this library; if not, write to the Free Software        */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  */
/*                                                                            */
/*                                                                            */
/* Contact Information:                                                       */
/*                                                                            */
/* Going Virtual, LLC                                                         */
/* 10800 Independence Pointe Pkwy                                             */
/* Matthews, NC 28105                                                         */
/* www.goingv.com                                                             */
/*                                                                            */
/* Robert E. Tierney                                                          */
/* bob.tierney@goingv.com                                                     */
/* 1 (704) 849-0731 x109                                                      */
/*                                                                            */
/******************************************************************************/
/*
 * $Id: gvibBindParm.i,v 1.7 2000/08/18 14:42:32 btierney Exp $
 * $Revision: 1.7 $
 */

%{  /* Start of SWIG-Free Section */

/******************************************************************************/
/*                                                                            */
/* bindParmVarying                                                            */
/*                                                                            */
/* This function binds the incoming data to a database VARCHAR column. As is  */
/* commonly done, the first two bytes of the data indicate how long the text  */
/* data actually is.                                                          */
/*                                                                            */
/******************************************************************************/
static short
bindParmVarying (XSQLVAR *var, PyObject *data, long index)
{
  char  *str_value;
  short *vaddr;
  short  str_length;

  if (!PyString_Check(data))
  {
    setBadParmException(data, "VARCHAR", index);
    return -1;
  }

  var->sqldata = (char *) Py_Malloc(var->sqllen + sizeof(short));
  if (!var->sqldata)
    return -1;

  str_value  = PyString_AsString(data);
  str_length = PyString_Size(data);
  if (var->sqllen < str_length)
    str_length = var->sqllen;

   vaddr = (short *) var->sqldata;
  *vaddr = str_length;

  memcpy(var->sqldata + sizeof(short), str_value, str_length);

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmText                                                               */
/*                                                                            */
/* This function is used to bind the incoming data to a database CHAR column. */
/*                                                                            */
/******************************************************************************/
static short
bindParmText (XSQLVAR *var, PyObject *data, long index)
{
  char  *str_value;
  short  str_length;

  if (!PyString_Check(data))
  {
    setBadParmException(data, "CHAR", index);
    return -1;
  }

  var->sqldata = (char *) Py_Malloc(var->sqllen);
  if (!var->sqldata)
    return -1;

  str_value  = PyString_AsString(data);
  str_length = PyString_Size(data);
  if (var->sqllen < str_length)
    str_length = var->sqllen;

  memset(var->sqldata, ' ', var->sqllen);
  memcpy(var->sqldata, str_value, str_length);

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmShort                                                              */
/*                                                                            */
/* This function is used to bind the incoming data to a database SMALLINT     */
/* column.  However the semantics of the column may be a NUMERIC or DECIMAL   */
/* type.  If so, call a different function to bind the data.                  */
/*                                                                            */
/******************************************************************************/
static short
bindParmShort (XSQLVAR *var, PyObject *data, long index)
{
  short rc;

  var->sqldata = (char *) Py_Malloc(sizeof(short));
  if (!var->sqldata)
    return -1;

  /*--------------------------------------------------------------------------*/
  /* If the subtype is greater than zero then the actual datatype is either   */
  /* NUMERIC or DECIMAL.                                                      */
  /*--------------------------------------------------------------------------*/
  if (var->sqlsubtype > 0)
  {
    rc = bindParmFixedShort(var, data, index);
    return rc;
  }

  if (!PyInt_Check(data))
  {
    setBadParmException(data, "SHORT", index);
    return -1;
  }

  *((short *) var->sqldata) = (short) PyInt_AsLong(data);

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmLong                                                               */
/*                                                                            */
/* This function is used to bind the incoming data to a database LONG column. */
/* However the semantics of the column may actually be NUMERIC or DECIMAL     */
/* types.                                                                     */
/*                                                                            */
/******************************************************************************/
static short
bindParmLong (XSQLVAR *var, PyObject *data, long index)
{
  short rc;

  var->sqldata = (char *) Py_Malloc(sizeof(long));
  if (!var->sqldata)
    return -1;

  /*--------------------------------------------------------------------------*/
  /* If the subtype is greater than zero then the actual datatype is either   */
  /* NUMERIC or DECIMAL.                                                      */
  /*--------------------------------------------------------------------------*/
  if (var->sqlsubtype > 0)
  {
    rc = bindParmFixedLong(var, data, index);
    return rc;
  }

  if (PyInt_Check(data))
  {
    *((long *) var->sqldata) = PyInt_AsLong(data);
  }
  else if (PyLong_Check(data))
  {
    *((long *) var->sqldata) = PyLong_AsLong(data);
  }
  else
  {
    setBadParmException(data, "LONG", index);
    return -1;
  }

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmFloat                                                              */
/*                                                                            */
/* This function is used to bind the incoming data to a database FLOAT column.*/
/*                                                                            */
/******************************************************************************/
static short
bindParmFloat (XSQLVAR *var, PyObject *data, long index)
{
  if (!PyFloat_Check(data))
  {
    setBadParmException(data, "FLOAT", index);
    return -1;
  }

  var->sqldata = (char *) Py_Malloc(sizeof(float));
  if (!var->sqldata)
    return -1;

  *((float *) var->sqldata) = (float) PyFloat_AsDouble(data);

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmDouble                                                             */
/*                                                                            */
/* This function is used to bind the incoming data to a database DOUBLE       */
/* column.                                                                    */
/*                                                                            */
/******************************************************************************/
static short
bindParmDouble (XSQLVAR *var, PyObject *data, long index)
{
  if (!PyFloat_Check(data))
  {
    setBadParmException(data, "DOUBLE", index);
    return -1;
  }

  var->sqldata = (char *) Py_Malloc(sizeof(double));
  if (!var->sqldata)
    return -1;

  *((double *) var->sqldata) = PyFloat_AsDouble(data);

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmTimestamp                                                          */
/*                                                                            */
/* This function is used to bind the incoming data to a database TIMESTAMP    */
/* column.                                                                    */
/*                                                                            */
/******************************************************************************/
static short
bindParmTimestamp (XSQLVAR *var, PyObject *data, long index)
{
  int    rc;
  long   precision = ISC_TIME_SECONDS_PRECISION;
  long   subseconds;
  int    yy, mm, dd, hh, mi;
  float  ss;
  struct tm dtime;

  rc = PyArg_ParseTuple(data, "iiiiif", &yy, &mm, &dd, &hh, &mi, &ss);
  if (!rc)
  {
    setBadParmException(data, "TIMESTAMP", index);
    return -1;
  }

  /*--------------------------------------------------------------------------*/
  /* Here we add a fudge factor to correct for floating point imprecision.    */
  /* For example the user pass 34.5486 but this may actually be 34.548599, so */
  /* we add 0.00001 (1.0 / (ISC_TIME_SECONDS_PRECISION * 10))                 */
  /*--------------------------------------------------------------------------*/
  ss += 1.0 / (precision * 10);

  var->sqldata = (char *) Py_Malloc(sizeof(ISC_TIMESTAMP));
  if (!var->sqldata)
    return -1;

  /*--------------------------------------------------------------------------*/
  /* Get the date and time into a 'tm' structure.  The seconds are            */
  /* truncated so the subseconds are lost here.  Encode and store the         */
  /* value.                                                                   */
  /*--------------------------------------------------------------------------*/
  dtime.tm_year = yy - 1900;
  dtime.tm_mon  = mm - 1;
  dtime.tm_mday = dd;
  dtime.tm_hour = hh;
  dtime.tm_min  = mi;
  dtime.tm_sec  = (int) ss;

  isc_encode_timestamp(&dtime, (ISC_TIMESTAMP *) var->sqldata);

  /*--------------------------------------------------------------------------*/
  /* Calculate the original subseconds value (as an integer) and add this     */
  /* to the value stored by the encode function.                              */
  /*                                                                          */
  /* Example:  A = seconds                =   30.7896 (encode stores 30)      */
  /*           B = seconds - dtime.tm_sec =    0.7896                         */
  /*           C = B * precision          = 7896.0000 (precision = 10000)     */
  /*  subseconds = C + (1 / precision)    = 7896                              */
  /*                                                                          */
  /* The value (1 / precision) is added to C to account for the fact that     */
  /* C may be 7896.0000 or it may be 7895.9999                                */
  /*                                                                          */
  /*--------------------------------------------------------------------------*/
  subseconds = ((ss - dtime.tm_sec) * precision) + (1.0 / precision);

  ((ISC_TIMESTAMP *) var->sqldata)->timestamp_time += subseconds;

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmTime                                                               */
/*                                                                            */
/* This function binds the incoming data to a database TIME column.           */
/*                                                                            */
/******************************************************************************/
static short
bindParmTime (XSQLVAR *var, PyObject *data, long index)
{
  int    rc;
  int    yy, mm, dd, hh, mi, ss;
  struct tm dtime;

  rc = PyArg_ParseTuple(data, "iiiiii", &yy, &mm, &dd, &hh, &mi, &ss);
  if (!rc)
  {
    setBadParmException(data, "TIME", index);
    return -1;
  }

  var->sqldata = (char *) Py_Malloc(sizeof(ISC_TIME));
  if (!var->sqldata)
    return -1;

  dtime.tm_year = yy - 1900;
  dtime.tm_mon  = mm - 1;
  dtime.tm_mday = dd;
  dtime.tm_hour = hh;
  dtime.tm_min  = mi;
  dtime.tm_sec  = ss;

  isc_encode_sql_time(&dtime, (ISC_TIME *) var->sqldata);

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmDate                                                               */
/*                                                                            */
/* This function binds the incoming data to a database DATE column.           */
/*                                                                            */
/******************************************************************************/
static short
bindParmDate (XSQLVAR *var, PyObject *data, long index)
{
  int    rc;
  struct tm dtime;
  int    yy, mm, dd, hh, mi, ss;

  rc = PyArg_ParseTuple(data, "iiiiii", &yy, &mm, &dd, &hh, &mi, &ss);
  if (!rc)
  {
    setBadParmException(data, "DATE", index);
    return -1;
  }

  var->sqldata = (char *) Py_Malloc(sizeof(ISC_DATE));
  if (!var->sqldata)
    return -1;

  dtime.tm_year = yy - 1900;
  dtime.tm_mon  = mm - 1;
  dtime.tm_mday = dd;
  dtime.tm_hour = hh;
  dtime.tm_min  = mi;
  dtime.tm_sec  = ss;

  isc_encode_sql_date(&dtime, (ISC_DATE *) var->sqldata);

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmInt64                                                              */
/*                                                                            */
/* This function is used to bind the incoming data to a database INT64        */
/* column.  However the semantics of the column may be NUMERIC or DECIMAL     */
/* type.                                                                      */
/*                                                                            */
/******************************************************************************/
static short
bindParmInt64 (XSQLVAR *var, PyObject *data, long index)
{
  short rc;

  var->sqldata = (char *) Py_Malloc(sizeof(ISC_INT64));
  if (!var->sqldata)
    return -1;

  /*--------------------------------------------------------------------------*/
  /* If the subtype is greater than zero then the actual datatype is either   */
  /* NUMERIC or DECIMAL.                                                      */
  /*--------------------------------------------------------------------------*/
  if (var->sqlsubtype > 0)
  {
    rc = bindParmFixedInt64(var, data, index);
    return rc;
  }
  
  if (PyInt_Check(data))
  {
    *((ISC_INT64 *) var->sqldata) = (ISC_INT64) PyInt_AsLong(data);
  }
  else if (PyLong_Check(data))
  {
    *((ISC_INT64 *) var->sqldata) = (ISC_INT64) PyLong_AsLong(data);
  }
  else
  {
    setBadParmException(data, "INT64", index);
    return -1;
  }
  
  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmBlob                                                               */
/*                                                                            */
/* This function is used to bind the incoming data to a database BLOB         */
/* column.  Note that only the QUAD value is stored in the database column.   */
/* The BLOB itself is stored elsewhere in the database.                       */
/*                                                                            */
/******************************************************************************/
static short
bindParmBlob (XSQLVAR *var, PyObject *data, long index, Cursor *self)
{
  if (!PyString_Check(data) && !PyBuffer_Check(data))
  {
    setBadParmException(data, "BLOB", index);
    return -1;
  }

  var->sqldata = (char *) Py_Malloc(sizeof(ISC_QUAD));
  if (!var->sqldata)
    return -1;

  /*--------------------------------------------------------------------------*/
  /* Call the putBlob function to create and write the blob data. It will     */
  /* store the new QUAD value in var->sqldata.                                */
  /*--------------------------------------------------------------------------*/
  if (putBlob(self, (ISC_QUAD *) var->sqldata, data) < 0)
    return -1;

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmFixedShort                                                         */
/*                                                                            */
/* This function implements the semantics of storing NUMERIC or DECIMAL data  */
/* into a database SMALLINT column.  This method will accept floating point   */
/* or integer data as input.                                                  */
/*                                                                            */
/******************************************************************************/
static short
bindParmFixedShort (XSQLVAR *var, PyObject *data, long index)
{
  int i;
  short  x, y, z;
  double multiplier = 1;
  double factor = 1;
  ISC_INT64 big_value;

  /*--------------------------------------------------------------------------*/
  /* For floating point data we need to shift the fractional part of the      */
  /* number to the left of the decimal point.  This is done my multiplying    */
  /* the data by a value derived from the scale of the column.  The resulting */
  /* value to the left of the decimal point is what we will store.            */
  /*--------------------------------------------------------------------------*/

  /*--------------------------------------------------------------------------*/
  /* Scale is stored in the XSQLVAR structure as a negative number.           */
  /*--------------------------------------------------------------------------*/
  for (i = 0; i > var->sqlscale; i--) 
    multiplier *= 10;

  for (i = 0; i >= var->sqlscale; i--)
    factor /= 10;

  if (PyFloat_Check(data))
  {
    double value;

    /*------------------------------------------------------------------------*/
    /* Here we add a fudge factor to account for the imprecision that can     */
    /* creep into floating point arithmetic.  The factor is calculated to     */
    /* be one more place of scale than that specified for the actual column.  */
    /*                                                                        */
    /* For example:  DECIMAL(4,2)                                             */
    /*                                                                        */
    /*   Precision:  4                                                        */
    /*       Scale:  2                                                        */
    /*  Multiplier:  100                                                      */
    /*      Factor:  .001                                                     */
    /*                                                                        */
    /*  Data Input:  13.54                                                    */
    /*                                                                        */
    /*      Step 1:  13.54 + .0001 = 13.541 if data was 13.540000             */
    /*               13.54 + .0001 = 13.540 if data was 13.539999             */
    /*                                                                        */
    /*      Step 2:  13.54 * 100 = 1354.0 or 1354.1                           */
    /*                                                                        */
    /*      Step 3:  Truncate and store 1354                                  */
    /*                                                                        */
    /*------------------------------------------------------------------------*/
    value  = PyFloat_AsDouble(data);
    value += factor;

    /*------------------------------------------------------------------------*/
    /* Do a cheesy check for overflow.                                        */
    /*------------------------------------------------------------------------*/
    big_value = (ISC_INT64)(value * multiplier);
    if (big_value > SHRT_MAX)
    {
      setOverflowException(index);
      return -1;
    }

    /*------------------------------------------------------------------------*/
    /* Lastly, truncate and store the integral value.                         */
    /*------------------------------------------------------------------------*/
    *((short *) var->sqldata) = (short)(value * multiplier);
  }
  else if (PyInt_Check(data))
  {
    long value;

    value = PyInt_AsLong(data);

    big_value = (ISC_INT64)(value * multiplier);
    if (big_value > SHRT_MAX)
    {
      setOverflowException(index);
      return -1;
    }

    *((short *) var->sqldata) = (short)(value * multiplier);
  }
  else
  {
    if (var->sqlsubtype == 1)
      setBadParmException(data, "NUMERIC (Internal Type = short)", index);
    else
      setBadParmException(data, "DECIMAL (Internal Type = short)", index);

    return -1;
  }

  return 0;
}

/******************************************************************************/
/*                                                                            */
/* binParmFixedLong                                                           */
/*                                                                            */
/* This function implements the semantics of storing NUMERIC or DECIMAL data  */
/* into a database INTEGER column.  This method will accept floating point    */
/* or integer data as input.                                                  */
/*                                                                            */
/******************************************************************************/
static short
bindParmFixedLong (XSQLVAR *var, PyObject *data, long index)
{
  int i;
  double multiplier = 1;
  double factor = 1;
  ISC_INT64 big_value;

  /*--------------------------------------------------------------------------*/
  /* See the comments in the bindParmFixedShort function.                     */
  /*--------------------------------------------------------------------------*/

  for (i = 0; i > var->sqlscale; i--)
    multiplier *= 10;

  for (i = 0; i >= var->sqlscale; i--)
    factor /= 10;

  if (PyFloat_Check(data))
  {
    double value;

    value  = PyFloat_AsDouble(data);
    value += factor;

    big_value = (ISC_INT64)(value * multiplier);

    if (big_value > LONG_MAX)
    {
      setOverflowException(index);
      return -1;
    }

    *((long *) var->sqldata) = (long)(value * multiplier);
  }
  else if (PyInt_Check(data))
  {
    long value;

    value     = PyInt_AsLong(data);
    big_value = (ISC_INT64)(value * multiplier);

    if (big_value > LONG_MAX)
    {
      setOverflowException(index);
      return -1;
    }

    *((long *) var->sqldata) = (long)(value * multiplier);
  }
  else if (PyLong_Check(data))
  {
    long value;

    value     = PyLong_AsLong(data);
    big_value = (ISC_INT64)(value * multiplier);

    if (big_value > LONG_MAX)
    {
      setOverflowException(index);
      return -1;
    }

    *((long *) var->sqldata) = (long)(value * multiplier);
  }
  else
  {
    if (var->sqlsubtype == 1)
      setBadParmException(data, "NUMERIC (Internal Type = long)", index);
    else
      setBadParmException(data, "DECIMAL (Internal Type = long)", index);

    return -1;
  }

  return 0;
}


/******************************************************************************/
/*                                                                            */
/* bindParmFixedInt64                                                         */
/*                                                                            */
/* This function implements the semantics of storing NUMERIC or DECIMAL data  */
/* into a database INT64 column.  This method will accept floating point or   */
/* integer data as input.                                                     */
/*                                                                            */
/******************************************************************************/
static short
bindParmFixedInt64 (XSQLVAR *var, PyObject *data, long index)
{
  int i;
  double multiplier = 1;
  double factor = 1;

  for (i = 0; i > var->sqlscale; i--)
    multiplier *= 10;

  for (i = 0; i >= var->sqlscale; i--)
    factor /= 10;

  /*--------------------------------------------------------------------------*/
  /* See the comments in the bindParmFixedShort function.                     */
  /*                                                                          */
  /* NOTE: There is no overflow check here.                                   */
  /*--------------------------------------------------------------------------*/

  if (PyFloat_Check(data))
  {
    double value;

    value  = PyFloat_AsDouble(data);
    value += factor;

    *((ISC_INT64 *) var->sqldata) = (ISC_INT64)(value * multiplier);
  }
  else if (PyInt_Check(data))
  {
    long value;

    value = PyInt_AsLong(data);

    *((ISC_INT64 *) var->sqldata) = (ISC_INT64)(value * multiplier);
  }
  else if (PyLong_Check(data))
  {
    long value;

    value = PyLong_AsLongLong(data);

    *((ISC_INT64 *) var->sqldata) = (ISC_INT64)(value * multiplier);
  }
  else
  {
    if (var->sqlsubtype == 1)
      setBadParmException(data, "NUMERIC (Internal Type = int64)", index);
    else
      setBadParmException(data, "DECIMAL (Internal Type = int64)", index);

    return -1;
  }

  return 0;
}


%}  /* End of SWIG-Free Section */
